/*
 * $Header: /cvs/lusers/lusers.c,v 1.17 2001/04/30 15:00:46 richard Exp $
 *
 *    lusers - user list program
 *    Copyright (C) 2000 Richard Kettlewell
 *
 *    This program is free software; you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation; either version 2 of the License, or
 *    (at your option) any later version.
 *
 *    This program is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with this program; if not, write to the Free Software
 *    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */

#include <time.h>
#include <unistd.h>
#include <string.h>
#include <stdio.h>
#include <errno.h>
#include <stdarg.h>
#include <stdlib.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/ioctl.h>
#include <fcntl.h>
#include <limits.h>
#include <termios.h>
#include <dirent.h>
#include <assert.h>
#include <fnmatch.h>
#include <ctype.h>
#include <signal.h>
#include <utmp.h>

#include "utmpfns.h"

#ifndef __GNUC__
# define __attribute__(x)
#endif

#if HAVE_UT_NAME && ! HAVE_UT_USER
# define ut_user ut_name
#endif

#if HAVE_LIBKVM
#include <kvm.h>
#include <sys/param.h>
#include <sys/sysctl.h>
#include <sys/user.h>

static kvm_t *kd;
#endif

#define RCFILE ".lusers"

struct sorttab {
  const char *name;
  int (*function)(const void *, const void *);
};

static enum {
  timestyle_sensible,
  timestyle_stupid,
} timestyle = timestyle_sensible;

static struct login {
  char *name;			/* username */
  size_t nlines;
  struct line *lines;		/* logins of this user */
} *users;
static size_t nusers;

struct line {
  char *line;			/* tty name */
  char *host;			/* hostname */
  char *cmd;			/* command */
  int msgable;			/* msgable? */
  int idle;			/* idle time */
  pid_t pid;			/* master PID */
};

static char **boring;		/* list of boring commands */
static size_t nboring;		/* number of boring commands */

static unsigned width = 80;
static unsigned xpos;

static int show_tty = 1;
static int show_command = 1;
static int show_idle = 1;
static int show_msgable = 1;
static int show_unmsgable = 0;
static int show_pts = 0;

static const char *sortspec_ttys = "+msgable+idle";
static const char *sortspec_users = "+name";

static void __attribute__((noreturn)) fatal(const char *fmt, ...);
static int do_compare(const void *a, const void *b,
		      const char *sortspec,
		      const struct sorttab *sorttab,
		      size_t nsorttab);
static int compare_line_msgable(const void *aa, const void *bb);
static int compare_line_idle(const void *aa, const void *bb);
static int compare_line(const void *aa, const void *bb);
static void sort_lines(struct login *u);
static int compare_user_name(const void *aa, const void *bb);
static int compare_user(const void *aa, const void *bb);
static void sort_users(void);
static void add_boring(const char *prog);
static void config_boring(const char *path,
			  int line,
			  char **argv);
static void config_timestyle(const char *path, int line, char **argv);
static void config_show(const char *path, int line, char **argv);
static void config_sort(const char *path, int line, char **argv);
static void configure(const char *path, int line, int argc, char **argv);
static void readconfig(void);
static void outputc(int c);
static void output(const char *s);
static void outputs(int n);
static struct login *getuser(const char *user);
static struct line *getline(struct login *u, const char *line);
static int stattty(const char *line, struct stat *sb);
static void register_login(const char *user,
			   pid_t pid,
			   const char *host,
			   void  *address,
			   const char *line);
#if HAVE_LINUX_PROCFS
static pid_t get_pgid_from_pid(pid_t pid);
#endif
#if HAVE_LIBKVM
static dev_t get_tty_number(const char *line);
#endif
#if HAVE_UT_PID
static int process_exists(pid_t pid);
#endif
static void scan_utmp(void);
static void display_username(const struct login *u);
static char *safe_strcat(char *buffer, size_t bufsize, const char *str);
static void display_line(const struct login *u, const struct line *l);
static void display_user(const struct login *u);
static void version(void);
static void usage(void);

static void __attribute__((noreturn)) fatal(const char *fmt, ...) {
  va_list ap;
  int e = errno;

  va_start(ap, fmt);
  vfprintf(stderr, fmt, ap);
  va_end(ap);
  if(*fmt && fmt[strlen(fmt)-1] == ':')
    fprintf(stderr, " %s\n", strerror(e));
  else
    fputc('\n', stderr);
  exit(1);
}

/* do a sort */

static int do_compare(const void *a, const void *b,
		      const char *sortspec,
		      const struct sorttab *sorttab,
		      size_t nsorttab) {
  const char *p;

  p = sortspec;
  while(*p) {
    int sign;
    size_t n, len;
    int result;
    
    if(*p == '+')
      sign = 1;
    else if(*p == '-')
      sign = -1;
    else
      return -1;
    ++p;
    for(len = 0; p[len] && isalnum((unsigned char)p[len]); ++len)
      ;
    for(n = 0; n < nsorttab; ++n)
      if(strlen(sorttab[n].name) == len
	 && !strncasecmp(sorttab[n].name, p, len))
	break;
    if(n >= nsorttab)
      return -1;
    if(a && b) {
      if((result = (*sorttab[n].function)(a, b)))
	return result * sign;
    }
    p += len;
  }
  return 0;
}

static int compare_line_msgable(const void *aa, const void *bb) {
  const struct line *a = aa;
  const struct line *b = bb;

  return b->msgable - a->msgable;
}

static int compare_line_idle(const void *aa, const void *bb) {
  const struct line *a = aa;
  const struct line *b = bb;

  return a->idle - b->idle;
}

static const struct sorttab sorttab_line[] = {
  {"msgable", compare_line_msgable},
  {"idle", compare_line_idle},
};

#define SORTTAB_LINE (sizeof sorttab_line / sizeof *sorttab_line)

/* sort the lines of a user; msg'able first, within msg'able least
   idle first */

static int compare_line(const void *aa, const void *bb) {
  return do_compare(aa, bb, sortspec_ttys,
		    sorttab_line, SORTTAB_LINE);
}

static void sort_lines(struct login *u) {
  qsort(u->lines, u->nlines, sizeof (struct line), compare_line);
}

static int compare_user_name(const void *aa, const void *bb) {
  const struct login *a = aa;
  const struct login *b = bb;
  
  return strcmp(a->name, b->name);
}

static const struct sorttab sorttab_users[] = {
  {"name", compare_user_name},
};

#define SORTTAB_USERS (sizeof sorttab_users / sizeof *sorttab_users)

/* sort users; currently alphabetically */

static int compare_user(const void *aa, const void *bb) {
  return do_compare(aa, bb, sortspec_users,
		    sorttab_users, SORTTAB_USERS);
}

static void sort_users(void) {
  size_t n;

  for(n = 0; n < nusers; n++)
    sort_lines(&users[n]);
  qsort(users, nusers, sizeof (struct login), compare_user);
}

static void add_boring(const char *prog) {
  if(!(boring = realloc(boring, sizeof (char *) * (1 + nboring))))
    fatal("error calling realloc:");
  if(!(boring[nboring++] = strdup(prog)))
    fatal("error calling strdup:");
}

static void config_boring(const char __attribute__((unused)) *path,
			  int __attribute__((unused)) line,
			  char **argv) {
  add_boring(*argv);
}

static void config_timestyle(const char *path, int line, char **argv) {
  if(!strcasecmp(*argv, "sensible"))
    timestyle = timestyle_sensible;
  else if(!strcasecmp(*argv, "stupid"))
    timestyle = timestyle_stupid;
  else
    fatal("%s:%d: unknown time style '%s'", path, line, *argv);
}

static const struct {
  const char *keyword;
  int *flag;
} showables[] = {
  {"tty", &show_tty},
  {"idle", &show_idle},
  {"command", &show_command},
  {"msgable", &show_msgable},
  {"unmsgable", &show_unmsgable},
  {"pts", &show_pts},
};

#define SHOWABLES (sizeof showables / sizeof *showables)

static void config_show(const char *path, int line, char **argv) {
  size_t n;

  for(n = 0; n < SHOWABLES && strcasecmp(showables[n].keyword, *argv); n++)
    ;
  if(n >= SHOWABLES)
    fatal("%s:%d: unknown showable '%s'", path, line, *argv);
  if(!strcasecmp(argv[1], "yes"))
    *showables[n].flag = 1;
  else if(!strcasecmp(argv[1], "no"))
    *showables[n].flag = 0;
  else
    fatal("%s:%d: showable status must be 'yes' or 'no', not '%s'",
	  path, line, argv[1]);
}

static const struct {
  const char *keyword;
  const char **strp;
  const struct sorttab *sorttab;
  size_t nsorttab;
} sortables[] = {
  {"ttys", &sortspec_ttys, sorttab_line, SORTTAB_LINE},
  {"users", &sortspec_users, sorttab_users, SORTTAB_USERS},
};

#define SORTABLES (sizeof sortables / sizeof *sortables)

static void config_sort(const char *path, int line, char **argv) {
  size_t n;

  for(n = 0; n < SORTABLES && strcasecmp(sortables[n].keyword, *argv); n++)
    ;
  if(n >= SORTABLES)
    fatal("%s:%d: unknown sortable '%s'", path, line, *argv);
  if(do_compare(0, 0, argv[1], sortables[n].sorttab, sortables[n].nsorttab))
    fatal("%s:%d: invalid sortspec '%s'", path, line, argv[1]);
  if(!(*sortables[n].strp = strdup(argv[1])))
    fatal("error calling strdup:");
}

static const struct {
  const char *opt;
  int args;
  void (*fn)(const char *path, int line, char **argv);
} options[] = {
  {"boring", 1, config_boring},
  {"timestyle", 1, config_timestyle},
  {"show", 2, config_show},
  {"sort", 2, config_sort},
};

#define OPTIONS (sizeof options / sizeof *options)

static void configure(const char *path, int line, int argc, char **argv) {
  while(argc-- > 0) {
    const char *opt = *argv++;
    size_t n;

    while(*opt == '-')
      ++opt;
    for(n = 0; n < OPTIONS && strcasecmp(opt, options[n].opt); ++n)
      ;
    if(n >= OPTIONS)
      fatal("%s:%d: unknown option '%s'", path, line, opt);
    if(argc < options[n].args)
      fatal("%s:%d: insufficient arguments to option '%s'", path, line, opt);
    (*options[n].fn)(path, line, argv);
    argc -= options[n].args;
    argv += options[n].args;
  }
}

static void readconfig(void) {
  char path[PATH_MAX];
  const char *home;
  FILE *fp;
  char line[4096];
  int lineno = 0;

  if(!(home = getenv("HOME")))
    fatal("HOME is not set");
  snprintf(path, sizeof path, "%s/%s", home, RCFILE);
  if(!(fp = fopen(path, "r"))) {
    if(errno == ENOENT)
      return;
    fatal("error opening %s:", path);
  }
  while(fgets(line, sizeof line, fp)) {
    char *p;
    char *argv[64];
    int argc;

    ++lineno;
    if(!strchr(line, '\n'))
      fatal("%s:%d: line too long", path, lineno);
    argc = 0;
    p = line;
    while(*(p += strspn(p, " \t\f\r\n"))) {
      char *q = p;
      
      if(argc >= 64)
	fatal("%s:%d: too many fields", path, lineno);
      argv[argc++] = p;
      if(*p == '"') {
	p++;
	while(*p && *p != '"') {
	  if(*p == '\\' && p[1])
	    ++p;
	  *q++ = *p++;
	}
	if(*p != '"')
	  fatal("%s:%d: unterminated string", path, lineno);
	*q = 0;
	++p;
      } else {
	p += strcspn(p, " \t\f\r\n");
	if(*p)
	  *p++ = 0;
      }
    }
    configure(path, lineno, argc, argv);
  }
  if(ferror(fp))
    fatal("error reading %s:", path);
  fclose(fp);
}

static void outputc(int c) {
  if(putchar(c) < 0)
    fatal("error writing to standard output:");
  if(c == '\n')
    xpos = 0;
  else
    ++xpos;
}

static void output(const char *s) {
  int c;

  while((c = *s++))
    outputc(c);
}

static void outputs(int n) {
  while(n-- > 0)
    outputc(' ');
}

static struct login *getuser(const char *user) {
  size_t n;

  for(n = 0; n < nusers && strcmp(user, users[n].name); n++)
    ;
  if(n < nusers)
    return &users[n];
  if(!(users = realloc(users, sizeof (struct login) * (nusers + 1))))
    fatal("error calling realloc:");
  memset(&users[nusers], 0, sizeof (struct login));
  if(!(users[nusers].name = strdup(user)))
    fatal("error calling strdup:");
  return &users[nusers++];
}

static struct line *getline(struct login *u, const char *line) {
  if(!(u->lines = realloc(u->lines, sizeof (struct line) * (u->nlines + 1))))
    fatal("error calling realloc:");
  memset(&u->lines[u->nlines], 0, sizeof (struct line));
  if(!(u->lines[u->nlines].line = strdup(line)))
    fatal("error calling strdup:");
  return &u->lines[u->nlines++];
}

static int stattty(const char *line, struct stat *sb) {
  char path[PATH_MAX];

  snprintf(path, sizeof path, "/dev/%s", line);
  if(stat(path, sb) < 0) {
    /* suppress error message for X logins */
    if(line[0] == ':')
      return 0;
    fprintf(stderr, "cannot stat %s: %s\n", path, strerror(errno));
    return -1;
  }
  return 0;
}

static void register_login(const char *user,
			   pid_t pid,
			   const char *host,
			   void __attribute__((unused)) *address,
			   const char *line) {
  struct login *u;
  struct line *l;
  struct stat sb;

  if(stattty(line, &sb))
    return;
  u = getuser(user);
  l = getline(u, line);
  if(!(l->host = strdup(host)))
    fatal("error calling strdup:");
  l->msgable = sb.st_mode & 0020 ? 1 : 0;
  l->idle = time(0) - sb.st_atime;
  l->pid = pid;
}

#if HAVE_LINUX_PROCFS
/* Finds the process group ID of the process which owns the tty that
   process PID is connected to */

static pid_t get_pgid_from_pid(pid_t pid) {
  char path[PATH_MAX];
  unsigned long pgrp;
  FILE *fp;
  
  snprintf(path, sizeof path, "/proc/%lu/stat", (unsigned long)pid);
  if(!(fp = fopen(path, "r"))) {
    fprintf(stderr, "error opening %s: %s\n", path, strerror(errno));
    return -1;
  }
  if(fscanf(fp, "%*u (%*[^)]) %*c %*u %*u %*u %*u %lu", &pgrp) != 1) {
    fprintf(stderr, "error reading %s: %s\n", path,
	    ferror(fp) ? strerror(errno) : "unexpected EOF");
    fclose(fp);
    return -1;
  }
  fclose(fp);
  return pgrp;
}
#endif

#if HAVE_LIBKVM
static dev_t get_tty_number(const char *line) {
  char path[PATH_MAX];
  struct stat sb;

  snprintf(path, sizeof path, "/dev/%s", line);
  if(stat(path, &sb) < 0)
    fatal("error calling stat %s:", path);
  return sb.st_rdev;
}
#endif

#if HAVE_UT_PID
static int process_exists(pid_t pid) {
  if(kill(pid, 0) < 0) {
    switch(errno) {
    case EPERM:
      return 1;
    case ESRCH:
      return 0;
    default:
      fatal("error checking for process %lu:", pid);
      break;
    }
  } else
    return 1;
}
#endif

static void scan_utmp(void) {
  struct utmp *u;
  
  setutent();
  while((errno = 0), (u = getutent())) {
#if HAVE_UT_TYPE && HAVE_UT_PID
    /* only interested in still-active user logins */
    if(u->ut_type != USER_PROCESS || !process_exists(u->ut_pid))
      continue;
#endif
    register_login(u->ut_user,
#if HAVE_UT_PID
		   u->ut_pid,
#else
		   0,
#endif
		   u->ut_host,
#if HAVE_UT_ADDR
		   &u->ut_addr,
#else
		   0,
#endif
		   u->ut_line);
  }
  if(errno)
    fatal("error reading utmp file:");
  endutent();
}

static void display_username(const struct login *u) {
  size_t s;
  
  if(strlen(u->name) > 8)
    s = 1;
  else
    s = 9 - strlen(u->name);
  output(u->name);
  outputs(s);
}

static char *safe_strcat(char *buffer, size_t bufsize, const char *str) {
  assert(strlen(buffer) + strlen(str) < bufsize);
  return strcat(buffer, str);
}

static void display_line(const struct login *u, const struct line *l) {
  char buffer[1024];

  buffer[0] = 0;
  if(show_msgable && l->msgable)
    safe_strcat(buffer, sizeof buffer, "*");
  else if(show_unmsgable && !l->msgable)
    safe_strcat(buffer, sizeof buffer, "*");
  if(show_tty) {
    char *ptr = l->line;
    if(!show_pts && !strncmp(ptr, "pts/", 4))
      ptr += 4;
    else if(!show_pts && !strncmp(ptr, "tty", 3))
      ptr += 3;
    safe_strcat(buffer, sizeof buffer, ptr);
  }
  if(show_command) {
    char cmdbuf[1024];
    pid_t pgrp;

#if HAVE_LINUX_PROCFS
    cmdbuf[0] = 0;
    if((pgrp = get_pgid_from_pid(l->pid)) != -1) {
      FILE *fp;
      char path[PATH_MAX];
      int c;

      snprintf(path, sizeof path, "/proc/%lu/stat", (unsigned long)pgrp);
      if((fp = fopen(path, "r"))) {
	size_t n = 0;

	/* XXX lazy /proc parsing */
	while((c = getc(fp)) != EOF && c != '(')
	  ;
	assert(c == '(');
	while((c = getc(fp)) != EOF && c != ')') {
	  if(n < sizeof cmdbuf - 1)
	    cmdbuf[n++] = c;
	}
	assert(c == ')');
	cmdbuf[n] = 0;
	fclose(fp);
      }
    }
#endif
#if HAVE_LIBKVM
    struct kinfo_proc *k;
    int count;

    cmdbuf[0] = 0;
    if(!(k = kvm_getprocs(kd, KERN_PROC_TTY, get_tty_number(l->line), &count)))
      fatal("error calling kvm_getprocs");
    if(count > 0) {
      int n;

      pgrp = k[0].kp_eproc.e_tpgid;
      for(n = 0; n < count; ++n)
	if(k[n].kp_proc.p_pid == pgrp)
	  break;
      if(n < count)
	safe_strcat(cmdbuf, sizeof cmdbuf, k[n].kp_proc.p_comm);
    }
#endif
    if(cmdbuf[0]) {
      size_t n;

      for(n = 0; n < nboring; n++)
	if(!fnmatch(boring[n], cmdbuf, 0))
	  break;
      if(n < nboring)
	cmdbuf[0] = 0;
    }

    if(cmdbuf[0]) {
      if(buffer[0])
	safe_strcat(buffer, sizeof buffer, ":");
      safe_strcat(buffer, sizeof buffer, cmdbuf);
    }
  }
  if(show_idle) {
    char idlebuf[1024];
    int idle = l->idle;

    switch(timestyle) {
    case timestyle_sensible:
      if(idle < 60)
	snprintf(idlebuf, sizeof idlebuf, "%ds", idle);
      else if(idle < 3600)
	snprintf(idlebuf, sizeof idlebuf, "%dm", idle / 60);
      else if(idle < 86400)
	snprintf(idlebuf, sizeof idlebuf, "%dh", idle / 3600);
      else if(idle < 604800)
	snprintf(idlebuf, sizeof idlebuf, "%dd", idle / 86400);
      else if(idle < 31556952)
	snprintf(idlebuf, sizeof idlebuf, "%dw", idle / 604800);
      else if(idle)
	snprintf(idlebuf, sizeof idlebuf, "%dy", idle / 31556952);
      break;
    case timestyle_stupid:
      if(idle < 1000)
	snprintf(idlebuf, sizeof idlebuf, "%ds", idle);
      else if(idle < 1000000)
	snprintf(idlebuf, sizeof idlebuf, "%dks", idle / 1000);
      else
	snprintf(idlebuf, sizeof idlebuf, "%dMs", idle / 1000000);
      break;
    }
    if(buffer[0])
      safe_strcat(buffer, sizeof buffer, " ");
    safe_strcat(buffer, sizeof buffer, idlebuf);
  }
  if(buffer[0]) {
    if(xpos + 1 + strlen(buffer) >= width) {
      output("\n");
      display_username(u);
    }
    output(" ");
    output(buffer);
  }
}

static void display_user(const struct login *u) {
  size_t i;
  
  display_username(u);
  for(i = 0; i < u->nlines; i++)
    display_line(u, &u->lines[i]);
  output("\n");
}

static void __attribute__((noreturn)) version(void) {
  if(printf("%s\n", VERSION) < 0
     || fclose(stdout) < 0)
    fatal("error writing to stdout:");
  exit(0);
}

static void __attribute__((noreturn)) usage(void) {
  if(printf("Usage:\n"
	    "\n"
	    "  lusers [options]\n"
	    "\n"
	    "  timestyle sensible|stupid    Set idle time display style\n"
	    "  boring <pattern>             Ignore boring commands\n"
	    "  show <option> yes|no         Set options\n"
	    "       command                 Show commands\n"
	    "       tty                     Show terminal names\n"
	    "       idle                    Show idle times\n"
	    "       msgable                 Show * for msg'able terminals\n"
	    "       unmsgable               Show * for un-msg'able terminals\n"
	    "  sort <list> <spec>           Set sort order\n"
	    "       users +name/-name       Sort users by name\n"
	    "       ttys +msgable/-msgable  Sort tty's by msg'ability\n"
	    "            +idle/-idle                ...by idle time\n"
	    "\n"
	    "Options may also be set in ~/.lusers\n"
	    "\n") < 0
     || fclose(stdout) < 0)
    fatal("error writing to stdout:");
  exit(0);
}

int main(int argc, char **argv) {
  size_t n;
  struct winsize w;

  if(argc == 2 && !strcmp(argv[1], "--help"))
    usage();
  if(argc == 2 && !strcmp(argv[1], "--version"))
    version();
  readconfig();
  configure("argv", 0, argc - 1, argv + 1);
#if HAVE_LIBKVM
  if(!(kd = kvm_open("/dev/null", "/dev/null", 0, O_RDONLY, argv[0])))
    return -1;
#endif
  /* pick up the width of the terminal */
  if(ioctl(1, TIOCGWINSZ, &w) < 0) {
    if(errno != ENOTTY && errno != EINVAL)
      fatal("error calling ioctl TIOCGWINSZ:");
  } else
    width = w.ws_col;
  /* find out what's happening */
  scan_utmp();
  /* sort people appropriately */
  sort_users();
  /* format it */
  for(n = 0; n < nusers; n++)
    display_user(&users[n]);
  if(fclose(stdout) < 0)
    fatal("error closing stdout:");
  return 0;
}

/*
Local Variables:
mode:c
c-basic-offset:2
comment-column:32
End:
*/
